"""
Q-value computation for RL model.

Author: Shirley Li (based on code by Nele Albers, https://doi.org/10.4121/22153898.v1)
Date: June 2024
"""
from copy import deepcopy
import itertools
import numpy as np

import reward_utils as util

def policy_evaluation(observation_space_size, discount_factor, trans_func,
                      reward_func, policy, q_vals, 
                      update_tolerance = 0.000000001):
    """
    Returns the q-values for a specific policy.
    
    Args:
        observation_space_size: number of observations
        discount_factor: discount factor of MDP
        trans_func: transition function
        reward_func: reward function
        policy: policy to compute q-values for
        update_tolerance: precision
    """
    q_vals_new = deepcopy(q_vals)
    num_act = len(trans_func[0])
    
    update = 1 # max. change in Q-value in one iteration
    while update > update_tolerance:
        update = 0
        for s in range(observation_space_size):
            for a in range(num_act):
                q_vals_new[s, a] = sum([trans_func[s, a, s_prime] * (discount_factor * q_vals[s_prime, int(policy[s_prime])] + reward_func[s, a, s_prime]) for s_prime in range(observation_space_size)])
                    
                update = max(update, abs(q_vals[s, a] - q_vals_new[s, a]))
        q_vals = deepcopy(q_vals_new)
        
    return q_vals_new
    

def get_Q_values_opt_policy(discount_factor, trans_func, reward_func):
    """
    Returns the Q-values for each state under the optimal policy as well
    as the optimal policy.
    
    Args:
        discount_factor: discount factor of MDP
        trans_func: transition function (dim.: num_states x num_actions x num_states)
        reward_func: reward function (dim.: num_states x num_actions)
        reward_dep_next_state (boolean, default: False): whether reward depends also on the next state
    """
    min_iterations = 100
    num_states = len(trans_func)
    num_act = len(trans_func[0])
    q_vals = np.zeros((num_states, num_act))

    policy = np.zeros(num_states)
    policy_new = np.ones(num_states)
    it = 0
    
    while (not np.array_equal(policy, policy_new)) or it < min_iterations:
        q_vals = policy_evaluation(num_states, discount_factor, 
                                   trans_func, reward_func, policy, q_vals)
        policy = policy_new
        policy_new = np.array([np.argmax(q_vals[s]) for s in range(num_states)])
        it += 1
    
    return q_vals, policy_new
    

def compute_q_vals_dynamics(data_in, weights, cost, num_feat_values, num_act = 2, 
                            discount_factor = 0.85, separate_rewards = False):
    """
    Compute Q-values and dynamics.

    Args:
        data_in (list): List with samples of the form <s0, s1, a, r>.
        weights (list): Weights for each reward.
        cost (float): The cost factor.
        feat_sel (list): Features to consider for policy computation.
        num_act (int): Number of possible actions.
        discount_factor (float): Discount factor for Q-value computation.
        separate_rewards (boolean): Whether we want to compute reward functions for effort and dropout separately.

    Returns:
        np-array: Q-values (num_states x num_actions)
        np-array: reward function (num_states x num_actions x num_states)
        np-array: transition function (num_states x num_actions x num_states)
        np-array: transition counts (num_states x num_actions x num_states)
        np-array: effort function (num_states x num_actions x num_states)
        np-array: dropout function (num_states x num_actions x num_states)

    """
    
    data = deepcopy(data_in)

    # All separate rewards
    list_of_rewards = list(np.array(data, dtype=object)[:, 3])

    # All actions
    list_of_actions = list(np.array(data, dtype=object)[:, 2])
    
    # Map effort responses to rewards from -1 to 1, with the mean mapped to 0.
    reward_list = util.get_weighted_rewards_list(list_of_rewards, weights, list_of_actions, cost)
    # now write these obtained reward values into "data" in place of the original rewards
    for i in range(len(reward_list)):
        data[i][3] = reward_list[i]
    
    num_samples = len(data)

    num_states = np.prod(num_feat_values)

    # Create list of all possible states
    abstract_states = [list(i) for i in itertools.product(*[list(range(f)) for f in num_feat_values])]
    
    # Compute transition function and reward function
    # Reward function is also dependent on next state
    trans_func = np.zeros((num_states, num_act, num_states))
    trans_counts = np.zeros((num_states, num_act, num_states))
    reward_func = np.zeros((num_states, num_act, num_states))
    reward_func_count = np.zeros((num_states, num_act, num_states))

    if separate_rewards:
        effort_func = np.zeros((num_states, num_act, num_states))
        dropout_func = np.zeros((num_states, num_act, num_states))

    for s_ind, s in enumerate(abstract_states):
        for data_index in range(num_samples):
            if data[data_index][0] == s:
                # index of next state
                next_state_index = abstract_states.index(data[data_index][1])
                
                trans_func[s_ind, data[data_index][2], next_state_index] += 1
                # Want to keep track of how often a transition occurs
                trans_counts[s_ind, data[data_index][2], next_state_index] += 1
                a = data[data_index][2]
                r = data[data_index][3]
                reward_func[s_ind, data[data_index][2], next_state_index] += r
                reward_func_count[s_ind, data[data_index][2], next_state_index] += 1

                if separate_rewards:
                    effort_func[s_ind, data[data_index][2], next_state_index] += list_of_rewards[data_index][0]
                    dropout_func[s_ind, data[data_index][2], next_state_index] += list_of_rewards[data_index][1]
       
        # Normalize reward and transition function
        for a in range(num_act):
            summed = sum(trans_func[s_ind, a])
            if summed > 0:
                trans_func[s_ind, a] /= summed
            # if we have no data on a state-action combination, we assume equal probability of transitioning to each other state
            else:
                trans_func[s_ind, a] = np.ones(num_states) / num_states
            for s_prime_ind in range(len(abstract_states)):
                if reward_func_count[s_ind, a, s_prime_ind] > 0:
                    reward_func[s_ind, a, s_prime_ind] /= reward_func_count[s_ind, a, s_prime_ind]
                    if separate_rewards:
                        effort_func[s_ind, a, s_prime_ind] /= reward_func_count[s_ind, a, s_prime_ind]
                        dropout_func[s_ind, a, s_prime_ind] /= reward_func_count[s_ind, a, s_prime_ind]
                    
    # Value iteration        
    q_values_exact, _ = get_Q_values_opt_policy(discount_factor, trans_func, 
                                                     reward_func)
    
    if separate_rewards:
        return q_values_exact, reward_func, effort_func, dropout_func, trans_func, trans_counts
    else:
        return q_values_exact, reward_func, trans_func, trans_counts
