"""
Feature selection based on the G-algorithm.

Author: Shirley Li (based on code by Nele Albers, https://doi.org/10.4121/22153898.v1)
Date: June 2024
"""
from copy import deepcopy
import itertools
import math
import numpy as np
import random
from scipy import stats

import reward_utils as util
import Compute_RL_QValues as qval

def approx_dynamics_blocks(data, num_blocks_plus_1, num_act, blocks, 
                           blocks_plus_1, f, feat_sel):
    """Approximate reward and transition functions for blocks.

    Args:
        data (list): List with samples of the form <s0, s1, a, r>.
        num_blocks_plus_1 (int): Number of blocks when using 1 more than the so far selected features.
        num_act (int): Number of actions.
        blocks (list): Abstract states based on so far selected features.
        blocks_plus_1 (list): Abstract states based on selecting one more feature.
        f (int): Candidate feature
        feat_sel (list): Already selected features
        
    Returns:
        np-array: reward function
        np-array: transition function

    """
    
    num_samples = len(data)
    
    trans_func = np.zeros((num_blocks_plus_1, num_act, num_blocks_plus_1))
    reward_func = np.zeros((num_blocks_plus_1, num_act, num_blocks_plus_1))
    reward_func_count = np.zeros((num_blocks_plus_1, num_act, num_blocks_plus_1))

    for b_ind, block in enumerate(blocks):  # for each block based on so far selected features
    
        for data_index in range(num_samples):  # for each data sample
            
            s = data[data_index][0][f]  # value for current feature 
            s_b = np.take(data[data_index][0], feat_sel)  # values for already selected features
            # the start state must be in the current block
            if list(s_b) == block:
                
                # next state and next state block
                s_prime = data[data_index][1][f]
                s_prime_b = np.take(data[data_index][1], feat_sel)
                # get the action and reward
                a = data[data_index][2]
                r = data[data_index][3]
                
                # update statistics
                trans_func[blocks_plus_1.index(list(s_b) + [s]), a, blocks_plus_1.index(list(s_prime_b) + [s_prime])] += 1
                reward_func[blocks_plus_1.index(list(s_b) + [s]), a, blocks_plus_1.index(list(s_prime_b) + [s_prime])] += r
                reward_func_count[blocks_plus_1.index(list(s_b) + [s]), a, blocks_plus_1.index(list(s_prime_b) + [s_prime])] += 1
    
    # Now need to normalize the reward and transition functions
    # For each block including the current candidate feature
    for b_ind in range(num_blocks_plus_1):  
        for a in range(num_act):  # for each action
            summed = sum(trans_func[b_ind, a])
            if summed > 0:
                trans_func[b_ind, a] /= summed
            # If we have no data on a state-action combination, we assume equal probability of transitioning to each other state
            else:
                trans_func[b_ind, a] = np.ones(num_blocks_plus_1) / num_blocks_plus_1
            for b_prime in range(num_blocks_plus_1):
                if reward_func_count[b_ind, a, b_prime] > 0:
                    reward_func[b_ind, a, b_prime] /= reward_func_count[b_ind, a, b_prime]
       
    return reward_func, trans_func


def approx_dynamics_single_feature(data, num_act, f, num_values):
    """Approximate reward and transition functions based on single features.

    Args:
        data (list): List with samples of the form <s0, s1, a, r>.
        num_act (int): Number of actions.
        f (int): feature
        num_values (int): Number of possible values for the feature.
        
    Returns:
        np-array: reward function (num_values x num_actions x num_values)
        np-array: transition function (num_values x num_actions x num_values)

    """
    num_samples = len(data)
    trans_func = np.zeros((num_values, num_act, num_values))
    reward_func = np.zeros((num_values, num_act, num_values))
    reward_func_count = np.zeros((num_values, num_act, num_values))
    for s in range(num_values):  # for each possible value of the feature
        for data_index in range(num_samples):  # for each data sample
            if data[data_index][0][f] == s:
                a = data[data_index][2]
                s_prime = data[data_index][1][f]  # feature value of next state
                r = data[data_index][3]
                
                # update statistics
                trans_func[s, a, s_prime] += 1
                reward_func[s, a, s_prime] += r
                reward_func_count[s, a, s_prime] += 1
   
        # Normalize transition and reward function for current value of feature
        for a in range(num_act):
            summed = sum(trans_func[s, a])
            if summed > 0:
                trans_func[s, a] /= summed
            # if we have no data on a state-action combination, we assume equal probability of transitioning to each other state
            else:
                trans_func[s, a] = np.ones(num_values) / num_values 
            for s_prime in range(num_values):
                if reward_func_count[s, a, s_prime] > 0:
                    reward_func[s, a, s_prime] /= reward_func_count[s, a, s_prime]
    
    return reward_func, trans_func

def feature_selection(data_in, all_states, weights, cost, feat_to_select, num_feat_values,
                      num_act = 2,
                      discount_factor = 0.85,
                      verbose=True):
    """Select state features in a way inspired by the G-algorithm based
    on Q-values.

    Args:
        data_in (list): List with samples of the form <s0, s1, a, r> where s0 and s1 are the original non-scaled values.
        weights (list): List of weights of the multiple rewards.
        feat_to_select (list): Candidate features.
        num_feat_to_select (int): Number of features to select.
        num_act (int): Number of possible actions.
        discount_factor (float): Discount factor for Q-value computation.

    Returns:
        list: Selected features.
        list: Criteria for selected features.
        list: New current state representations.
        list: New next state representations.
    """
    data = deepcopy(data_in)
    # All separate rewards and actions
    list_of_rewards = list(np.array(data, dtype=object)[:, 3])
    list_of_actions = list(np.array(data, dtype=object)[:, 2])

    # Lists of all current and next state values in their own scales
    list_of_s0 = list(np.array(data, dtype=object)[:, 0])
    list_of_s1 = list(np.array(data, dtype=object)[:, 1])
    
    # Get list of weighted rewards.
    reward_list = util.get_weighted_rewards_list(list_of_rewards, weights, list_of_actions, cost)
    # now write these obtained reward values into "data" in place of the original rewards
    for i in range(len(reward_list)):
        data[i][3] = reward_list[i]

    num_feat = len(feat_to_select)
    num_feat_to_select = len(num_feat_values)

    # Select first feature
    q_values = np.zeros((num_feat, num_feat_values[0], num_act))
    p_values = np.zeros((num_feat))

    # Number of values that the first feature can take
    num_values_first = num_feat_values[0]

    # Get new state representations given the number of possible values that we want for the first feature (i.e. map every state value to binary or tertiary values)
    new_states, new_next_states = get_state_representation(all_states, list_of_s0, list_of_s1, num_values_first)
    for i in range(len(new_states)):
        data[i][0] = new_states[i]
        data[i][1] = new_next_states[i]
    
    for f in range(num_feat):
        # Compute transition function and reward function for feature
        reward_func, trans_func = approx_dynamics_single_feature(data,
                                                                 num_act, f, num_values_first)
    
        # Value iteration for current feature    
        q_values[f], _ = qval.get_Q_values_opt_policy(discount_factor, 
                                                      trans_func, 
                                                      reward_func)
        
        # ANOVA Test to get p-values to compare the Q-values of different values for the features
        p_values[f] = stats.f_oneway(*q_values[f])[1]
    
    if verbose:
        print("P-values for single features:", p_values)
    min_p_val = np.nanmin(p_values)  # minimum p-value for t-test, ignoring nan-values
    feat_sel_options = [i for i in range(num_feat) if p_values[i] == min_p_val]
    feat_sel = [random.choice(feat_sel_options)]  # choose randomly if there are multiple best features
    criterion = "First feature -> min. p-value: " + str(round(min_p_val, 4))
    if len(feat_sel_options) > 1:
        criterion += " random from " + str(feat_sel_options)
    feat_sel_criteria = [criterion]
    if verbose:
        print("First feature selected:", feat_sel[0], criterion)
    
    # Select remaining features
    for j in range(num_feat_to_select - 1):
        num_feat_sel = len(feat_sel)
        feat_not_sel = [f for f in range(num_feat) if not f in feat_sel]
        num_feat_not_sel = len(feat_not_sel)
        
        # all possible states based on the so far selected features
        blocks = [list(i) for i in itertools.product(*[list(range(num_feat_values[f])) for f in range(num_feat_sel)])]
        
        # all possible states based on the so far selected features plus the new feature
        blocks_plus_1 = [list(i) for i in itertools.product(*[list(range(num_feat_values[f])) for f in range(num_feat_sel + 1)])]
        num_blocks_plus_1 = len(blocks_plus_1)
        
        q_values_2 = np.zeros((num_feat_not_sel, num_blocks_plus_1, num_act))
        p_values_2 = np.zeros(num_feat_not_sel)

        # Number of values that the next feature can take
        num_values = num_feat_values[num_feat_sel]

        # Get new state representations given the number of possible values for the next feature
        new_states, new_next_states = get_state_representation(all_states, list_of_s0, list_of_s1, num_values)
        for i in range(len(new_states)):
            # If the feature is not yet selected, we replace the feature value in the state representation
            data[i][0] = [new_states[i][f] if f in feat_not_sel else data[i][0][f] for f in range(num_feat)]
            data[i][1] = [new_next_states[i][f] if f in feat_not_sel else data[i][1][f] for f in range(num_feat)]

        for f_ind, f in enumerate(feat_not_sel):  # for each not yet selected feature
            # Approximate reward and transition function
            reward_func, trans_func = approx_dynamics_blocks(data, 
                                                             num_blocks_plus_1, 
                                                             num_act, blocks, 
                                                             blocks_plus_1, f,
                                                             feat_sel)
        
            # Now we need to compute the Q-values
            # Value iteration for current feature    
            q_values_2[f_ind], _ = qval.get_Q_values_opt_policy(discount_factor, 
                                                                trans_func, 
                                                                reward_func)
            
            # Separate Q-values for the possible values of the feature
            q_values_all_blocks = []
            for n in range(num_values):
                q_values_all_blocks.append(np.array([q_values_2[f_ind][b_ind] for b_ind in range(n, len(blocks_plus_1), num_values)]).flatten())

            # Now that we have the data on all blocks, we run an ANOVA test to get the p-values
            p_values_2[f_ind] = stats.f_oneway(*q_values_all_blocks)[1]
        
        if verbose:
            print("P-values for features after selecting previous feature:", p_values_2)
        # Select next feature
        min_val_curr = min(p_values_2)
        feat_min_p_val = [i for i in range(num_feat_not_sel) if p_values_2[i] == min_val_curr]
        if len(feat_min_p_val) == 1:
            feat_sel.append(feat_not_sel[feat_min_p_val[0]])
            feat_sel_criteria.append("Min p-value: " + str(round(min_val_curr, 4)))
        # multiple best features -> choose one randomly
        else:
            feat_sel.append(feat_not_sel[random.choice(feat_min_p_val)])
            feat_sel_criteria.append("Min p-value: " + str(round(min_val_curr, 4)) + " random from " + str(feat_min_p_val))

        if verbose:   
            print("Feature selected:", feat_sel[-1])
            print("Criterion:", feat_sel_criteria[-1])
        
    s0 = [list(np.take(state, feat_sel)) for state in np.array(data, dtype=object)[:, 0]]  # current state in correct representation format
    s1 = [list(np.take(state, feat_sel)) for state in np.array(data, dtype=object)[:, 1]]  # next state in correct representation format
    return feat_sel, feat_sel_criteria, s0, s1

def get_single_state_representation(all_states, s0, s1, num_values):
    """
    Converts the given state values into a new format of a given number of possible values for a single state feature.

    Args:
        s0 (list): The current state.
        s1 (list): The next state.
        num_values (int): The number of possible values for each feature (can either be 2, 3 or 4).

    Returns:
        tuple: A tuple containing the new current state representation and the new next state representation.
    """
    new_state = []
    new_next_state = []

    if num_values == 4:
        first_quartile_cutoff = np.percentile(all_states, 25)
        second_quartile_cutoff = np.percentile(all_states, 50)
        third_quartile_cutoff = np.percentile(all_states, 75)
        new_state = [0 if x <= first_quartile_cutoff else 1 if first_quartile_cutoff < x <= second_quartile_cutoff else 2 if second_quartile_cutoff < x <= third_quartile_cutoff else 3 for x in s0]
        new_next_state = [0 if x <= first_quartile_cutoff else 1 if first_quartile_cutoff < x <= second_quartile_cutoff else 2 if second_quartile_cutoff < x <= third_quartile_cutoff else 3 for x in s1]
    if num_values == 3:
        first_tertile_cutoff = np.percentile(all_states, 100/3)
        second_tertile_cutoff = np.percentile(all_states, 200/3)
        new_state = [0 if x <= first_tertile_cutoff else 1 if first_tertile_cutoff < x <= second_tertile_cutoff else 2 for x in s0]
        new_next_state = [0 if x <= first_tertile_cutoff else 1 if first_tertile_cutoff < x <= second_tertile_cutoff else 2 for x in s1]
    if num_values == 2:
        cutoff = np.percentile(all_states, 50)
        new_state = [0 if x <= cutoff else 1 for x in s0]
        new_next_state = [0 if x <= cutoff else 1 for x in s1]
    return new_state, new_next_state

def get_state_representation(all_states, s0, s1, num_values):
    """
    Converts the given state values into a new format of a given number of possible values.

    Args:
        s0 (list): The current state.
        s1 (list): The next state.
        num_values (int): The number of possible values for each feature.

    Returns:
        tuple: A tuple containing the new current state representation and the new next state representation.
    """
    new_states = []
    new_next_states = []

    # Get representation of current and next state for each feature
    for i in range(len(s0[0])):
        list_of_s0 = list(np.array(s0, dtype=object)[:, i])
        list_of_s1 = list(np.array(s1, dtype=object)[:, i])
        list_of_all_states = list(np.array(all_states, dtype=object)[:, i])

        # Get representation of current and next state for single feature
        feature_representation = get_single_state_representation(list_of_all_states, list_of_s0, list_of_s1, num_values)
        new_states.append(feature_representation[0])
        new_next_states.append(feature_representation[1])

    new_states = [list(state) for state in zip(*new_states)]
    new_next_states = [list(next_state) for next_state in zip(*new_next_states)]
    return new_states, new_next_states
