"""
Utility functions for calculating rewards.

Author: Shirley Li

Date: June 2024
"""
import numpy as np

def get_rewards_separate_values(efforts, dropouts, effort_mean, dropout_mean):
    """
    Calculates rewards for efforts and dropouts separately based on the given means.

    Args:
        efforts (list): List of effort values.
        dropouts (list): List of dropout values.
        effort_mean (float): Mean value for efforts.
        dropout_mean (float): Mean value for dropouts.

    Returns:
        list: List of rewards for efforts and dropouts separately.

    """
    r_e = [get_reward(e, effort_mean) for e in efforts]
    r_d = [get_reward(d, dropout_mean) for d in dropouts]
    return [[e, d] for e, d in zip(r_e, r_d)]

def get_weighted_reward(rewards, weights, action, cost):
    """
    Calculates the weighted reward based on the given separate rewards, weights, action, and cost.
    
    Parameters:
        rewards (list): An array of rewards.
        weights (list):  Weights for each reward.
        action (int): The action taken.
        cost (float): The cost factor.
    
    Returns:
        float: The weighted reward.
    """
    return np.dot(rewards, weights) - (action * cost)

def get_weighted_rewards_list(rewards_list, weights, actions, cost):
    """
    Calculates the list of weighted rewards based on the given list of separate rewards, weights, actions, and cost.

    Args:
        rewards_list (list): A list of separate rewards.
        weights (list): Weights for each reward.
        actions (list): A list of actions corresponding to each reward.
        cost (float): The cost factor.

    Returns:
        list: A list of weighted rewards.
    """
    return [get_weighted_reward(rewards, weights, action, cost) for rewards, action in zip(rewards_list, actions)]

def get_reward(x, mean):
    """
    Calculates the reward based on the given value and mean.

    Parameters:
    x (float): The value for which the reward is calculated.
    mean (float): The mean value.

    Returns:
    float: The calculated reward.
    """
    if x == mean:
        return 0
    elif x > mean:
        return 1 - ((10-x) / (10-mean))
    else:
        return -1 + (x / mean)
    