"""
Data preprocessing to extract transition samples for the reinforcement learning model.

The raw data that is needed to run this:
- sessionsdata_bsc_anonym.csv

The processed data that is generated:
- all_states.csv
- rl_samples.csv

Author: Shirley Li
Date: June 2024
"""	

import pandas as pd
import numpy as np
import Analysis.reward_utils as utils

FILE_PATH = 'Data\sessionsdata_bsc_anonym.csv'
STATE_NAMES = ['state_humansupport', 'state_importance', 'state_selfefficacy', 'state_energy']

def get_state_list(data):
    """
    Returns a list of states extracted from the given data.

    Parameters:
    - data: The input data containing the states.

    Returns:
    - full_states: A list of states extracted from the data.
    """
    full_states = []
    for _, row in data.iterrows():
        full_state = [row[s] for s in STATE_NAMES]  # Get the state values
        full_states.append(full_state)
    return full_states

def get_next_states_rewards(data):
    """
    Get the next state and rewards for each sample from the given data.

    Parameters:
    - data: The input data containing the states and rewards.

    Returns:
    - next_full_states: A list of next states.
    - rewards: A list of rewards.

    """
    users = data.index.get_level_values('rand_id').unique()  # Get all person ids
    next_full_states, rewards = [], []
    for user in users:
        user_states = data.loc[user]['s0']  # Session states for this person
        for i in range(1, len(user_states)):
            # Get the next state and reward for the current state
            rewards.append(data.loc[user]['r'][i+1])
            next_full_states.append(user_states[i+1])
        # Add NaN values for the last state or if there is only one session for this person
        rewards.append(np.nan)
        next_full_states.append(np.nan)
    return next_full_states, rewards

if __name__ == "__main__":
    df = pd.read_csv(FILE_PATH)

    # Pivot the data to have the states as columns
    df = df.pivot_table(index=['rand_id', 'session_num'], columns='response_type', values='response_value', aggfunc=lambda x: x.iloc[0])

    # Remove unneccessary columns
    df = df[["dropout_response", "effort", "human_support_after_session", "state_humansupport", "state_importance", "state_selfefficacy", "state_energy"]]
    
    df = df.astype(float)
    df['s0'] = get_state_list(df)  # Get the state for each session
    df = df[~df['s0'].apply(lambda x: any(np.isnan(x)))]  # Remove samples with missing state values

    # Save the data on all states
    df.to_csv("Analysis/Data/all_states.csv")

    # Scale dropout response from [-5, 5] to [0, 10]
    df['dropout_response'] = df['dropout_response'].apply(lambda x: x + 5)

    # Calculate the mean effort and dropout response of only the samples that have both values
    means = df.dropna(subset=['effort', 'dropout_response'])
    effort_mean = means['effort'].mean()
    dropout_mean = means['dropout_response'].mean()

    # Calculate the rewards for each sample
    df['r'] = utils.get_rewards_separate_values(df['effort'], df['dropout_response'], effort_mean, dropout_mean)

    # Get the next state and reward for each sample
    df['s1'], df['r'] = get_next_states_rewards(df)

    # Remove samples with missing reward values
    df.dropna(inplace=True, subset=['r'])
    
    # Only keep relevant columns and rename them
    df = df[['s0', 's1', 'human_support_after_session', 'r']]	
    df.columns = ['s0', 's1', 'a', 'r']
    df['a'] = df['a'].astype(int)

    df.to_csv("Analysis/Data/rl_samples.csv")